// autogenerated by syzkaller (http://github.com/google/syzkaller)

#define _GNU_SOURCE

#include <arpa/inet.h>
#include <errno.h>
#include <fcntl.h>
#include <fcntl.h>
#include <linux/if.h>
#include <linux/if_ether.h>
#include <linux/if_tun.h>
#include <linux/ip.h>
#include <linux/tcp.h>
#include <net/if_arp.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/stat.h>
#include <sys/syscall.h>
#include <sys/uio.h>
#include <unistd.h>

__attribute__((noreturn)) static void doexit(int status)
{
  volatile unsigned i;
  syscall(__NR_exit_group, status);
  for (i = 0;; i++) {
  }
}

#include <stdint.h>
#include <string.h>

const int kFailStatus = 67;
const int kRetryStatus = 69;

static void fail(const char* msg, ...)
{
  int e = errno;
  va_list args;
  va_start(args, msg);
  vfprintf(stderr, msg, args);
  va_end(args);
  fprintf(stderr, " (errno %d)\n", e);
  doexit((e == ENOMEM || e == EAGAIN) ? kRetryStatus : kFailStatus);
}

static void vsnprintf_check(char* str, size_t size, const char* format,
                            va_list args)
{
  int rv;

  rv = vsnprintf(str, size, format, args);
  if (rv < 0)
    fail("tun: snprintf failed");
  if ((size_t)rv >= size)
    fail("tun: string '%s...' doesn't fit into buffer", str);
}

static void snprintf_check(char* str, size_t size, const char* format,
                           ...)
{
  va_list args;

  va_start(args, format);
  vsnprintf_check(str, size, format, args);
  va_end(args);
}

#define COMMAND_MAX_LEN 128
#define PATH_PREFIX                                                    \
  "PATH=/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin "
#define PATH_PREFIX_LEN (sizeof(PATH_PREFIX) - 1)

static void execute_command(const char* format, ...)
{
  va_list args;
  char command[PATH_PREFIX_LEN + COMMAND_MAX_LEN];
  int rv;

  va_start(args, format);
  memcpy(command, PATH_PREFIX, PATH_PREFIX_LEN);
  vsnprintf_check(command + PATH_PREFIX_LEN, COMMAND_MAX_LEN, format,
                  args);
  rv = system(command);
  if (rv != 0)
    fail("tun: command \"%s\" failed with code %d", &command[0], rv);

  va_end(args);
}

static int tunfd = -1;
static int tun_frags_enabled;

#define SYZ_TUN_MAX_PACKET_SIZE 1000

#define MAX_PIDS 32
#define ADDR_MAX_LEN 32

#define LOCAL_MAC "aa:aa:aa:aa:aa:%02hx"
#define REMOTE_MAC "bb:bb:bb:bb:bb:%02hx"

#define LOCAL_IPV4 "172.20.%d.170"
#define REMOTE_IPV4 "172.20.%d.187"

#define LOCAL_IPV6 "fe80::%02hxaa"
#define REMOTE_IPV6 "fe80::%02hxbb"

#define IFF_NAPI 0x0010
#define IFF_NAPI_FRAGS 0x0020

static void initialize_tun(uint64_t pid)
{
  if (pid >= MAX_PIDS)
    fail("tun: no more than %d executors", MAX_PIDS);
  int id = pid;

  tunfd = open("/dev/net/tun", O_RDWR | O_NONBLOCK);
  if (tunfd == -1) {
    printf(
        "tun: can't open /dev/net/tun: please enable CONFIG_TUN=y\n");
    printf("otherwise fuzzing or reproducing might not work as "
           "intended\n");
    return;
  }

  char iface[IFNAMSIZ];
  snprintf_check(iface, sizeof(iface), "syz%d", id);

  struct ifreq ifr;
  memset(&ifr, 0, sizeof(ifr));
  strncpy(ifr.ifr_name, iface, IFNAMSIZ);
  ifr.ifr_flags = IFF_TAP | IFF_NO_PI | IFF_NAPI | IFF_NAPI_FRAGS;
  if (ioctl(tunfd, TUNSETIFF, (void*)&ifr) < 0) {
    ifr.ifr_flags = IFF_TAP | IFF_NO_PI;
    if (ioctl(tunfd, TUNSETIFF, (void*)&ifr) < 0)
      fail("tun: ioctl(TUNSETIFF) failed");
  }
  if (ioctl(tunfd, TUNGETIFF, (void*)&ifr) < 0)
    fail("tun: ioctl(TUNGETIFF) failed");
  tun_frags_enabled = (ifr.ifr_flags & IFF_NAPI_FRAGS) != 0;

  char local_mac[ADDR_MAX_LEN];
  snprintf_check(local_mac, sizeof(local_mac), LOCAL_MAC, id);
  char remote_mac[ADDR_MAX_LEN];
  snprintf_check(remote_mac, sizeof(remote_mac), REMOTE_MAC, id);

  char local_ipv4[ADDR_MAX_LEN];
  snprintf_check(local_ipv4, sizeof(local_ipv4), LOCAL_IPV4, id);
  char remote_ipv4[ADDR_MAX_LEN];
  snprintf_check(remote_ipv4, sizeof(remote_ipv4), REMOTE_IPV4, id);

  char local_ipv6[ADDR_MAX_LEN];
  snprintf_check(local_ipv6, sizeof(local_ipv6), LOCAL_IPV6, id);
  char remote_ipv6[ADDR_MAX_LEN];
  snprintf_check(remote_ipv6, sizeof(remote_ipv6), REMOTE_IPV6, id);

  execute_command("sysctl -w net.ipv6.conf.%s.accept_dad=0", iface);

  execute_command("sysctl -w net.ipv6.conf.%s.router_solicitations=0",
                  iface);

  execute_command("ip link set dev %s address %s", iface, local_mac);
  execute_command("ip addr add %s/24 dev %s", local_ipv4, iface);
  execute_command("ip -6 addr add %s/120 dev %s", local_ipv6, iface);
  execute_command("ip neigh add %s lladdr %s dev %s nud permanent",
                  remote_ipv4, remote_mac, iface);
  execute_command("ip -6 neigh add %s lladdr %s dev %s nud permanent",
                  remote_ipv6, remote_mac, iface);
  execute_command("ip link set dev %s up", iface);
}

static void setup_tun(uint64_t pid, bool enable_tun)
{
  if (enable_tun)
    initialize_tun(pid);
}

static uintptr_t syz_open_dev(uintptr_t a0, uintptr_t a1, uintptr_t a2)
{
  if (a0 == 0xc || a0 == 0xb) {
    char buf[128];
    sprintf(buf, "/dev/%s/%d:%d", a0 == 0xc ? "char" : "block",
            (uint8_t)a1, (uint8_t)a2);
    return open(buf, O_RDWR, 0);
  } else {
    char buf[1024];
    char* hash;
    strncpy(buf, (char*)a0, sizeof(buf));
    buf[sizeof(buf) - 1] = 0;
    while ((hash = strchr(buf, '#'))) {
      *hash = '0' + (char)(a1 % 10);
      a1 /= 10;
    }
    return open(buf, a2, 0);
  }
}

long r[48];
void loop()
{
  memset(r, -1, sizeof(r));
  r[0] = syscall(__NR_mmap, 0x20000000ul, 0xfff000ul, 0x3ul, 0x32ul,
                 0xfffffffffffffffful, 0x0ul);
  memcpy((void*)0x20005000,
         "\x2f\x64\x65\x76\x2f\x62\x69\x6e\x64\x65\x72\x23\x00", 13);
  r[2] = syz_open_dev(0x20005000ul, 0x0ul, 0x0ul);
  memcpy((void*)0x201acff3,
         "\x2f\x64\x65\x76\x2f\x62\x69\x6e\x64\x65\x72\x23\x00", 13);
  r[4] = syz_open_dev(0x201acff3ul, 0x0ul, 0x0ul);
  r[5] = syscall(__NR_mmap, 0x20000000ul, 0x2000ul, 0x0ul, 0x20011ul,
                 r[4], 0x0ul);
  r[6] = syscall(__NR_ioctl, r[4], 0x40046207ul, 0x0ul);
  *(uint64_t*)0x20008fd0 = (uint64_t)0x4;
  *(uint64_t*)0x20008fd8 = (uint64_t)0x0;
  *(uint64_t*)0x20008fe0 = (uint64_t)0x20003fe8;
  *(uint64_t*)0x20008fe8 = (uint64_t)0x0;
  *(uint64_t*)0x20008ff0 = (uint64_t)0x0;
  *(uint64_t*)0x20008ff8 = (uint64_t)0x20009000;
  *(uint32_t*)0x20003fe8 = (uint32_t)0x630b;
  r[14] = syscall(__NR_ioctl, r[4], 0xc0306201ul, 0x20008fd0ul);
  *(uint64_t*)0x20007000 = (uint64_t)0x4;
  *(uint64_t*)0x20007008 = (uint64_t)0x0;
  *(uint64_t*)0x20007010 = (uint64_t)0x20005fd4;
  *(uint64_t*)0x20007018 = (uint64_t)0x0;
  *(uint64_t*)0x20007020 = (uint64_t)0x0;
  *(uint64_t*)0x20007028 = (uint64_t)0x20002000;
  *(uint32_t*)0x20005fd4 = (uint32_t)0x40486311;
  *(uint64_t*)0x20005fd8 = (uint64_t)0x0;
  *(uint64_t*)0x20005fe0 = (uint64_t)0x0;
  r[24] = syscall(__NR_ioctl, r[2], 0xc0306201ul, 0x20007000ul);
  *(uint64_t*)0x2000dfd0 = (uint64_t)0x54;
  *(uint64_t*)0x2000dfd8 = (uint64_t)0x0;
  *(uint64_t*)0x2000dfe0 = (uint64_t)0x2000dfac;
  *(uint64_t*)0x2000dfe8 = (uint64_t)0xca;
  *(uint64_t*)0x2000dff0 = (uint64_t)0x0;
  *(uint64_t*)0x2000dff8 = (uint64_t)0x2000df36;
  *(uint32_t*)0x2000dfac = (uint32_t)0x400c630e;
  *(uint32_t*)0x2000dfb0 = (uint32_t)0x3;
  *(uint64_t*)0x2000dfb4 = (uint64_t)0x4;
  *(uint32_t*)0x2000dfbc = (uint32_t)0x40106308;
  *(uint64_t*)0x2000dfc0 = (uint64_t)0x0;
  *(uint64_t*)0x2000dfc8 = (uint64_t)0x3;
  *(uint32_t*)0x2000dfd0 = (uint32_t)0x400c630e;
  *(uint32_t*)0x2000dfd4 = (uint32_t)0x4;
  *(uint64_t*)0x2000dfd8 = (uint64_t)0x0;
  *(uint32_t*)0x2000dfe0 = (uint32_t)0x40046306;
  *(uint32_t*)0x2000dfe4 = (uint32_t)0x1;
  *(uint32_t*)0x2000dfe8 = (uint32_t)0x40086303;
  *(uint64_t*)0x2000dfec = (uint64_t)0x0;
  *(uint32_t*)0x2000dff4 = (uint32_t)0x40086310;
  *(uint64_t*)0x2000dff8 = (uint64_t)0x4;
  memcpy((void*)0x2000df36,
         "\x12\x25\xbd\x0c\xe6\xff\xb6\xfa\x93\xcf\x99\x08\x03\xad\x59"
         "\x00\x79\x04\xcf\x4b\x9e\x3e\x9f\x75\x1e\xc4\x7a\xc3\xd7\x8c"
         "\x46\x3c\x71\xd1\xcc\x33\x2c\xbe\x58\xac\x10\xd8\xf7\xd8\x04"
         "\xb3\xa7\x1f\xe6\xa4\x26\x46\x11\x42\x56\xe3\x8e\x1a\x44\x7e"
         "\x77\xee\x7b\x5b\x5f\x1c\x97\xe0\x0b\x8c\x88\x25\xdc\x1c\x49"
         "\xe0\xdd\xb3\x8c\x1c\x4d\xeb\xf8\x0f\xe1\x3a\x47\xbe\xb3\x01"
         "\xfb\xc4\x8a\x73\xff\x16\x59\xef\x42\x50\xe4\xf3\x1c\xfd\x1b"
         "\x0c\x21\xe1\xa8\x7d\xaa\xc3\x17\xe8\x2f\x47\x2d\x45\x05\xfc"
         "\x50\x46\x58\x00\x4b\x79\x93\xde\x11\x23\x3f\x60\xa9\x0e\x1c"
         "\xe0\x34\xb8\xa8\xca\x17\x68\x54\xe9\x3f\xc8\xab\x9d\x3a\x17"
         "\x6b\x31\x76\xce\xaa\x62\x64\x44\x01\x69\xda\x3e\x1c\x39\x1b"
         "\x95\x92\xda\x7d\xa3\x5c\x2d\xa5\x5d\xe5\xe7\x68\x15\x0f\x51"
         "\x35\x72\xbd\x99\x10\x3a\x2c\x99\xaf\xc2\x07\xd1\x22\xe6\x29"
         "\x11\xa1\xc8\x90\xd1\x4a\xa9",
         202);
  r[47] = syscall(__NR_ioctl, r[4], 0xc0306201ul, 0x2000dfd0ul);
}

int main()
{
  setup_tun(0, true);
  loop();
  return 0;
}
